/*
 * Ratiometric Frequency Counter, LED version
 *
 * displays 6 digits from 1Hz up to ~40MHz
 * cycle time 4M/F_CPU ... 16M/F_CPU
 * theoretical lower frequency limit: F_CPU/16M
 * theoretical upper frequency limit: 4*F_CPU
 *
 * (c) Axel (XL) Schwenke, axel.schwenke@gmx.net
 *
 * $Id: main.c 163 2012-10-15 09:08:45Z schwenke $
 */

#define MINCYCLES (4*1024*1024L)  /* min number of fref cycles */
#define SLEEP     1               /* sleep in wait loops */
#define WDT       1               /* enable hardware watchdog */

#if F_CPU > (16*1024*1024-512)
#error "F_CPU is too big! Must fit in 24 bits"
#endif

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/sleep.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <util/delay.h>

#include "arith.h"
#include "ports.h"


/* constants */

const uint8_t segmentcode[]=
{
    SegA | SegB | SegC | SegD | SegE | SegF       , /* 0 */
           SegB | SegC                            , /* 1 */
    SegA | SegB        | SegD | SegE        | SegG, /* 2 */
    SegA | SegB | SegC | SegD               | SegG, /* 3 */
           SegB | SegC               | SegF | SegG, /* 4 */
    SegA        | SegC | SegD        | SegF | SegG, /* 5 */
    SegA        | SegC | SegD | SegE | SegF | SegG, /* 6 */
    SegA | SegB | SegC                            , /* 7 */
    SegA | SegB | SegC | SegD | SegE | SegF | SegG, /* 8 */
    SegA | SegB | SegC | SegD        | SegF | SegG, /* 9 */
    SegA | SegB | SegC        | SegE | SegF | SegG, /* A */
                  SegC | SegD | SegE | SegF | SegG, /* b */
                         SegD | SegE        | SegG, /* c */
           SegB | SegC | SegD | SegE        | SegG, /* d */
    SegA               | SegD | SegE | SegF | SegG, /* E */
    SegA                      | SegE | SegF | SegG  /* F */
};


/* function prototypes */

void hw_init(void);
void calculate_frequency(void);
void calculate_period(void);


/* global variables */

volatile union {    /* 24 bit counter f_x */
    uint8_t  b[3];
} nx;

volatile union {    /* 24 bit counter f_ref */
    uint16_t w[1];
    uint8_t  b[3];
} nref;

volatile union {    /* 48 bit result */
    uint32_t q[1];
    uint8_t  b[6];
} res;

uint32_t fref;      /* reference frequency */
uint8_t  jumpers;   /* jumper settings */

uint8_t buffer[7]=  /* display buffer */
{
    /* initialized to show version information xx.yy.zz */
    SegA | SegB | SegC | SegD | SegE | SegF               , /* 0  */
           SegB | SegC                             | SegDP, /* 1. */
    SegA | SegB | SegC | SegD | SegE | SegF               , /* 0  */
    SegA        | SegC | SegD | SegE | SegF | SegG | SegDP, /* 6. */
    SegA | SegB | SegC | SegD | SegE | SegF               , /* 0  */
    SegA | SegB | SegC | SegD | SegE | SegF               , /* 0  */
    /* turn on all extra LEDs */
    SegA | SegB | SegC | SegD | SegE | SegF | SegG | SegDP
};

/* constants in EEPROM */

EEMEM uint32_t ee_fref = F_CPU;


/* the code */

int __attribute__((OS_main)) main(void)
{
    hw_init();
    _delay_ms(1000);

    while (1) {

        /* clear counters */
        clear_low();
        nx.b[2]= 0;
        TCNT0= 0;
        clear_high();

        TCCR1B &= ~_BV(ICES1);  /* capture falling edge */
        TIMSK  |=  _BV(TOIE1);  /* enable Timer1 overflow irq */
        sei();
        start_high();
        while (gate_is_closed()) {
            wdt_reset();
            sleep_cpu();
        }

        while ((nref.b[2] < (MINCYCLES/65536)+1) && (nx.b[2] < 0x7c)) {
            wdt_reset();
            sleep_cpu();
        }

        TCCR1B |=  _BV(ICES1);  /* capture rising edge */
        start_low();
        while (nref.b[2] && gate_is_open()) {
            wdt_reset();
        }

        cli();

        if (nref.b[2] == 0) {
            buffer[6] |= U_OVR;

        } else {

            /* save display and switch to read mode */
            uint8_t tmpdig= digit_read();
            uint8_t tmpseg= data_read();
            digits_off();
            data_input();

            /* read fx counter */
            lat_oe_low();
            lat_oe_high();      /* rising edge latches 74HC590 */
            lat_oe_low();       /* low enables 74HC590 drivers */

            register union {    /* ugly workaround */
                uint16_t w;     /* to avoid even uglier gcc code */
                uint8_t  b[2];
            } tmp;

            tmp.b[0]= TCNT0;
            tmp.b[1]= nx.b[2];
            tmp.w <<= 1;
            tmp.b[0] |= fx9_read();

            nx.b[0]= data_read();
            nx.b[1]= tmp.b[0];
            nx.b[2]= tmp.b[1];

            /* read jumper/switch settings */
            lat_oe_high();
            data_write(0xff);
            data_output();
            data_input();
            _delay_us(10);
            jumpers= data_read();

            /* restore display and continue multiplexing */
            data_output();
            data_write(tmpseg);
            digit_on(tmpdig);
            sei();

            /* fref = EEPROM constant + calibration */
            fref= eeprom_read_dword(&ee_fref);
            register uint32_t cal= ADC;
            if (cal != 0) {
                fref += (cal - 512);
            }

            /* calculate result and write to display buffer */
            if (jumper_is_set(JUMPER_MODE)) {
                calculate_period();
            } else {
                calculate_frequency();
            }

        }
    }
}


/* Interrupt Service Routines */

/* Timer 0 Overflow - f_x counter */
ISR(TIMER0_OVF_vect)
{
    ++nx.b[2];
}

/* Timer 1 Overflow - f_ref counter */
ISR(TIMER1_OVF_vect)
{
    ++nref.b[2];
}

/* Timer 1 Capture - gate open/close */
ISR(TIMER1_CAPT_vect)
{
    if ((TCCR1B & _BV(ICES1)) == 0) {
        /* falling edge = gate opens */
        nref.w[0]= ICR1;
        nref.b[2]= 0;
    } else {
        /* rising edge = gate closes */
        register uint16_t tmp= ICR1;
        if ((TIFR & _BV(TOV1)) && tmp < 256) {
            /* overflow irq pending - possible collision */
            ++nref.b[2];
        }
        if (tmp < nref.w[0]) {
            --nref.b[2];
        }
        nref.w[0]= tmp - nref.w[0];
        TIMSK &= ~_BV(TOIE1);  /* disable and */
        TIFR   =  _BV(TOV1);   /* clear overflow interrupt */
    }
}

/* Timer 2 Overflow - display multiplex */
ISR(TIMER2_OVF_vect)
{
    static uint8_t digit= 0;

    digit = (digit == 6 ? 0 : digit+1);

    digits_off();
    data_write(buffer[digit] ^ SegInversion);
    digit_on(digit);
}


/* Sub Routines */

void hw_init(void)
{
    /* pullups enabled on all inputs */
    /* data bus pins are outputs */
    /* START=L, CLR=L, LAT/OE=H */

    PORTB= ~(_BV(CLEAR_PIN) | _BV(START_PIN));
    DDRB=   (_BV(CLEAR_PIN) | _BV(START_PIN));

    PORTC= ~(_BV(ANALOG_PIN));
    DDRC=  0x1F;

    PORTD= 0xFF;
    DDRD=   (0x07 | _BV(LAT_OE_PIN) | 0xE0);

    /* disable Analog Comparator */

    ACSR = (_BV(ACD));

    /* ADC senses PC5 continuously, Vref=Aref */

    ADMUX=  (_BV(MUX2) | _BV(MUX0));
    ADCSRA= (_BV(ADEN) | _BV(ADSC) | _BV(ADFR) | _BV(ADPS2) | _BV(ADPS1) | _BV(ADPS0));

    /* Timer 0: ext. trigger, falling edge */
    /* Timer 1: system clock */
    /* Timer 2: system clock / 64 -> ~125Hz multiplex rate */

    TCCR0=  (_BV(CS02) | _BV(CS01));
    TCCR1A= 0;
    TCCR1B= (_BV(CS10));
    TCCR2=  (_BV(CS22));
    ASSR=   0;

    /* Timer 0: capture irq and
     * Timer 1: overflow irq and
     * Timer 2: overflow irq enabled */

    TIMSK=  (_BV(TICIE1) | _BV(TOIE0) | _BV(TOIE2));

#if SLEEP
    set_sleep_mode(SLEEP_MODE_IDLE);
    sleep_enable();
#endif

#if WDT
    wdt_enable(WDTO_2S);
#endif
}


void calculate_frequency(void)
{
    uint8_t unit=U_Hz;
    uint8_t decimal=7; /* decimal point after digit #7 */

    /* res *= fref * nx */
    res.q[0] = fref;
    mymul((void*) &res, (void*) &nx);

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF8) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= nref */
    mydiv((void*) &res, (void*) &nref);

    /* res = PackedBCD(res) */
    mybin2bcd((void*) &res);

    /* skip leading zeros */
    while (decimal && (res.b[3] & 0xF0) == 0) {
        res.q[0] <<= 4;
        --decimal;
    }

    /* adjust unit prefix */
    if (decimal>2) {
        unit     = U_kHz;
        decimal -= 3;
    }
    if (decimal>2) {
        unit     = U_MHz;
        decimal -= 3;
    }

    /* write 6 digits */
    for (register uint8_t i=0; i<6; i++) {
        buffer[i]= segmentcode[(res.b[3] & 0xF0) >> 4];
        if (i==decimal) {
            buffer[i] |= SegDP;
        }
        res.q[0] <<= 4;
    }

    /* append unit */
    buffer[6]= unit;

}


void calculate_period(void)
{
    uint8_t unit=U_ns;
    uint8_t decimal=16; /* decimal point after digit #16 */

    union {
        uint8_t  b[4];
        uint32_t q[1];
    } nx2;

    /* normalize nx */
    nx2.b[0] = nx.b[0];
    nx2.b[1] = nx.b[1];
    nx2.b[2] = nx.b[2];
    nx2.b[3] = 0;

    while ((nx2.b[2] & 0xF8) == 0) {
        nx2.q[0] *= 10;
        ++decimal;
    }

    /* res = nref */
    res.b[0] = nref.b[0];
    res.b[1] = nref.b[1];
    res.b[2] = nref.b[2];
    res.b[3] = 0;
    res.b[4] = 0;
    res.b[5] = 0;

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF0) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= nx */
    mydiv((void*) &res, (void*) &nx2);

    /* multiply by 10 until big enough
     * this shifts the decimal position */
    while ((res.b[5] & 0xF0) == 0 && decimal > 1) {
        /* res *= 10 */
        mymulby10((void*) &res);
        --decimal;
    }

    /* res /= fref */
    mydiv((void*) &res, (void*) &fref);

    /* res = PackedBCD(res) */
    mybin2bcd((void*) &res);

    /* skip leading zeros */
    while (decimal && (res.b[3] & 0xF0) == 0) {
        res.q[0] <<= 4;
        --decimal;
    }

    /* adjust unit prefix */
    if (decimal>2) {
        unit     = U_us;
        decimal -= 3;
    }
    if (decimal>2) {
        unit     = U_ms;
        decimal -= 3;
    }

    /* write 6 digits */
    for (register uint8_t i=0; i<6; i++) {
        buffer[i]= segmentcode[(res.b[3] & 0xF0) >> 4];
        if (i==decimal) {
            buffer[i] |= SegDP;
        }
        res.q[0] <<= 4;
    }

    /* append unit */
    buffer[6]= unit;

}


